#include <ESP8266WiFi.h>
#include <EEPROM.h>

// Servidor al port TCP 80
WiFiServer server(80);

// Variables de connexió de la xarxa WIFI
String ssid = "";
String password = "";
IPAddress ip;

IPAddress gateway(192, 168, 1, 1);
IPAddress subnet(255, 255, 255, 0);

// Reset per hardware, el pin D1 està connectat al pin RST
void reset() {
  digitalWrite(D1, LOW);
}

// Iniciem el mòdul
void setup() {
  // Led que s'il·lumina quan es reben dades
  pinMode(2, OUTPUT); digitalWrite(2, HIGH);
  // Pin de reset en valor HIGH
  pinMode(D1, OUTPUT); digitalWrite(D1, HIGH);
  // Iniciar el mòdul WIFI
  Serial.println("Iniciado módulo WIFI");
  // Iniciem el serial a 19200 bauds
  Serial.begin(19200);
  // Llegim la configuració de l'EEPROM
  readConfig();
  // Si tenim configuració i aconsegueix connectar
  if ( ssid.length() > 1 ) {
    // Si hem connectar amb el WIFI sortim
    if (modeWIFI()) return;
  }
  // Si no aconsegueix connectar es posa en mode AP per ser configurat
  modoAP();
}

// Llegeix la configuració de l'eeprom
void readConfig() {
  EEPROM.begin(512); // Iniciem la memòria per llegir la configuració
  delay(10);

  // 32 bytes pel nom de la xarxa
  for (int i = 0; i < 32; ++i) {
    ssid += char(EEPROM.read(i));
  }
  Serial.print("SSID : ");
  Serial.println(ssid);
  // 32 bytes pel password
  for (int i = 32; i < 64; ++i) {
    password += char(EEPROM.read(i));
  }
  Serial.print("password : ");
  Serial.println(password);

  // La direcció IP és un byte per cada part de l'adreça 255.255.255.255
  byte aux[4];
  for (int i = 0; i < 4; ++i) {
    aux[i] = EEPROM.read(64 + i);
  }
  ip = IPAddress(aux);
  Serial.print("ip : ");
  Serial.println(ip);

  Serial.print("Gateway : ");
  Serial.println(gateway);
  Serial.print("Subnet : ");
  Serial.println(subnet);
}

// Escriure a l'eeprom, s'ha de fer en el mateix ordre i mides 
void writeConfig() {
  EEPROM.begin(512); // Iniciem la memòria per escriure la configuració
  delay(10);
  for (int i = 0; i < 32; ++i) {
    if (i < ssid.length())  EEPROM.write(i, ssid[i]);
    else EEPROM.write(i, 0);
  }
  for (int i = 0; i < 32; ++i) {
    if (i < password.length()) EEPROM.write(32 + i, password[i]);
    else EEPROM.write(32 + i, 0);
  }
  for (int i = 0; i < 4; ++i) {
    EEPROM.write(64 + i, ip[i]);
  }
  EEPROM.commit();
}

// Funció que connectar amb la xarxa WIFI i retorna TRUE si ho aconsegueix
bool modeWIFI(void) {
  // Si no hi ha configuració retorna false
  if (ip[0] == 255) return false;
  
  WiFi.mode(WIFI_STA);
  WiFi.config(ip, gateway, subnet);

  Serial.print("Connectant a ");
  Serial.println(ssid);
  WiFi.begin(ssid, password);
  // Fa diversos intents
  for (int i = 0; i < 20; i++) {
    if (WiFi.status() == WL_CONNECTED) {
      Serial.println("WiFi connected.");
      Serial.println("IP address: ");
      Serial.println(WiFi.localIP());
      server.begin();
      return true;
    }
    delay(500);
  }
  // Si no pot retorna false
  return false;
}

// Inicar el mode AP, crea una WIFI pròpia, la comunicació es produeix en el socket TCP 80
bool modoAP(void) {
  bool init = false;
  while (true) {
    // Si no estem en mode
    if (!init) {
      init = true;
      // Iniciar la configuració del mode AP
      Serial.println("Modo AP");
      WiFi.mode(WIFI_AP);
      WiFi.disconnect();
      delay(100);
      // Indica el nom de la xarxa i la paraula de pas
      WiFi.softAP("MyRobot", "12345678", 6);
      IPAddress softIP(192, 168, 1, 1);
      IPAddress softMASK(255, 255, 255, 0);
      WiFi.softAPConfig(softIP, softIP, softMASK);
      // Inicia el server
      server.begin();
    }

    // Mira si es connecta algun client
    WiFiClient client = server.available();
    if (client) {
      String message = "";
      while (client.connected()) {
        if (client.available()) {
          // Si tenim client i dades
          char c = client.read();
          int ini, fin;
          String aux;
          // Es pinta pel serial per debug
          Serial.print(c);
          switch (c) {
            // El caràcter de principi de paquet és un 0x02 STX
            case 0x02 :
              message = "";
              break;
            // EL caràcter per indicar fi de paquet és un 0x03 ETX
            case 0x03 :
              // Enviem un OK al mòbil conforme s'ha rebut bé la configuració
              for (int i = 0; i < 5; i++) {
                client.println("OK");
                delay(10);
              }
              // Els paràmetres estan separats per el caràcter ~
              Serial.println(message);
              ini = 0;
              fin = message.indexOf("~");
              // Llegim el SID
              ssid = message.substring(ini, fin);
              Serial.print("SSID : ");
              Serial.println(ssid);
              
              // Llegir la paraula de pas
              ini = fin + 1;
              fin = message.indexOf("~", ini);
              password = message.substring(ini, fin);
              Serial.print("password : ");
              Serial.println(password);

              // Llegir la IP que tindrà el robot
              ini = fin + 1;
              fin = message.indexOf("~", ini);
              aux = message.substring(ini, fin);
              ip.fromString(aux);
              Serial.print("ip : ");
              Serial.println(aux);

              // Si aconsegueix connectar, emmagatzema la configuració i reseteja el programa
              if (modeWIFI()) {
                writeConfig();
                reset();
                return true; // El return no caldria perquè farà el reset, però per si de cas surt de la funció
              }
              init = false;
              break;
            default:
              // Acumula els caràcters llegits
              message += c;
              break;
          }
        }
      }
      // Para el client
      client.stop();
    }
  }
}

void loop() {
  // Si arriba al loop és que ha aconseguit connectar a la xarxa wifi configurada
  WiFiClient client = server.available();   // Espera que es connecti un client

  if (client) {                             // Client operatiu
    digitalWrite(2, LOW);
    // Mentre estigui connectat
    String http = "";
    String command = "";

    // Espera que hi hagin dades
    while(!client.available())  delay(1);
    // Si ja hi ha dades les llegim
    while (client.available()) {
      // Llegim un byte i el retrasmetem cap al port serie
      char c = client.read();
      http += c;
      if (c == '$') break;  // Si rebem el caràcter $ indica final de l'ordre
      if (c == '@' || command.indexOf("@") >= 0) command += c; // El caràcter @ indica inici de l'ordre
    }
    // Buidem la resta de caràcters que no ens interessan ja que el navegador envia amb el protocol HTTP
    // diverses dades que no fem servir
    while (client.available()) {
      client.read();
    }
    // Si la variable http comença per GET indica que l'ordre bé des d'un navegador i hem de retornar la resposta
    // segons el protocol HTTP
    if (http.startsWith("GET ")) {
      client.print("HTTP/1.1 200 OK\r\n");  // Indiquem que la petició és correcta
      client.print("Access-Control-Allow-Origin: *\r\n"); // Això permet fer crides des del server disseny-apli.cat
      client.print("Content-Type: text/html\r\n"); // Indiquem que la resposta és de text
      client.print("\r\n");  
      client.print("<!DOCTYPE HTML>\r\n<html>\r\nOK</html>\r\n");  // Retornem un OK
    } else {
      client.println("OK"); // Si no és un navegador i és l'APP amb un OK n'hi ha prou
    }
    // Si tenim una ordre l'enviem cap a l'Arduino
    if (command.startsWith("@")) {
      Serial.print(command);
      Serial.println();
    }
    // Apaguem el led que indica les transmissions i tanquem el client
    digitalWrite(2, HIGH);
    client.stop();
  }
  // Si no estem connectats al wifi reiniciem el programa perque ho torni a provar
  if (WiFi.status() != WL_CONNECTED) reset();
}
